/*
   Copyright (C) 1998,1999 T. Scott Dattalo

This file is part of the libgpsim library of gpsim

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, see 
<http://www.gnu.org/licenses/lgpl-2.1.html>.
*/


/*
  packages.h

Here's where all of the pic packages are defined

 */


#ifndef __PIC_PACKAGES_H__
#define __PIC_PACKAGES_H__

#include "gpsim_classes.h"
#include "pic-processor.h"
#include "stimuli.h"
#include "packages.h"



#endif // __PIC_PACKAGES_H__
