/*
   Copyright (C) 1998 T. Scott Dattalo

This file is part of the libgpsim library of gpsim

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, see 
<http://www.gnu.org/licenses/lgpl-2.1.html>.
*/

/*
 * gpsim_def.h
 *
 * This include file contains global definitions.
 */

#ifndef __GPSIM_VERSION_H__
#define __GPSIM_VERSION_H__

//#define GPSIM_VERSION VERSION 

#define MAX_PROGRAM_MEMORY 0xffff

#define DEFAULT_GUI_UPDATE_RATE   0x100000

#define MAX_HLL_FILES 20

#endif // __GPSIM_VERSION_H__
